
#ifndef PLS_INTERNAL_DATA_STRUCTURES_DELAYED_INITIALIZATION_WRAPPER_H_
#define PLS_INTERNAL_DATA_STRUCTURES_DELAYED_INITIALIZATION_WRAPPER_H_

#include <array>
#include <utility>

#include "pls/internal/base/error_handling.h"

namespace pls {
namespace internal {
namespace data_structures {

/**
 * Allows to reserve space for an uninitialized member variable.
 * The member must be initialized before usage using the provided
 * perfect forwarding constructor method.
 *
 * Makes sure to call the wrapped objects de-constructor when an object is wrapped.
 */
template<typename T>
class delayed_initialization_wrapper {
 public:
  delayed_initialization_wrapper() : memory_{}, initialized_{false} {}
  template<typename ...ARGS>
  explicit delayed_initialization_wrapper(ARGS &&...args): memory_{}, initialized_{true} {
    new(memory_) T(std::forward<ARGS>(args)...);
  }

  ~delayed_initialization_wrapper() {
    if (initialized_) {
      memory_->~T();
    }
  }

  template<typename ...ARGS>
  void initialize(ARGS &&...args) {
    PLS_ASSERT(initialized_, "Can only initialize delayed wrapper object once!")

    new(memory_) T(std::forward<ARGS>(args)...);
    initialized_ = true;
  }
  T &object() {
    PLS_ASSERT(initialized_, "Can not use an uninitialized delayed wrapper object!")
    return *reinterpret_cast<T *>(memory_);
  }

 private:
  std::array<char, sizeof(T)> memory_;
  bool initialized_;
};

}
}
}

#endif // PLS_INTERNAL_DATA_STRUCTURES_DELAYED_INITIALIZATION_WRAPPER_H_
