
#ifndef PLS_INTERNAL_SCHEDULING_CONTINUATION_H_
#define PLS_INTERNAL_SCHEDULING_CONTINUATION_H_

#include <type_traits>
#include <atomic>
#include <utility>

#include "pls/internal/data_structures/delayed_initialization_wrapper.h"

namespace pls {
namespace internal {
namespace scheduling {

class base_continuation {
 public:
  virtual void run() = 0;
  virtual ~base_continuation() = 0;
};

class continuation_node {
 public:

 private:
  // Pointer to memory region reserved for the companion continuation.
  // Must be a buffer big enough to hold any continuation encountered in the program.
  base_continuation *continuation_;

  // Linked list property of continuations (continuation chains as memory management).
  // Each continuation knows its chain start to allow stealing a whole chain in O(1)
  // without the need to traverse back to the chain start.
  continuation_node *cont_chain_start_;
  continuation_node *prev_, *next_;

  // When blocked on this continuation, we need to know what other chain we
  // got offered by the stealing thread.
  // For this we need only the head of the other chain (as each continuation is a
  // self describing entity for its chain up to the given node).
  continuation_node *offered_chain_;
};

template<typename R1, typename R2, typename F>
class continuation : public base_continuation {
 public:
  void run() override {
    // TODO: integrate this better into the runtime system.
    //       E.g. handle passing the result to the parent continuation
    function_.object()(result_1_.object(), result_2_.object());
  }
  ~continuation() override = default;

  template<typename R1ARG>
  void store_result_1_(R1ARG &&result_1) {
    static_assert(std::is_same<R1, typename std::decay<R1ARG>::type>::value,
                  "must only copy/move objects in, not construct them");
    result_1_.initialize(std::forward<R1ARG>(result_1));
  }

  template<typename R2ARG>
  void store_result_2(R2ARG &&result_1) {
    static_assert(std::is_same<R2, typename std::decay<R2ARG>::type>::value,
                  "must only copy/move objects in, not construct them");
    result_2_.initialize(std::forward<R2ARG>(result_1));
  }

  template<typename FARG>
  void store_function(FARG &&function) {
    static_assert(std::is_same<F, typename std::decay<FARG>::type>::value,
                  "must only copy/move objects in, not construct them");
    function_.initialize(function);
  }

 private:
  // We plan to only init the members for a continuation on the slow path.
  // If we can execute everything inline we simply skip it saving runtime overhead.
  template<typename T>
  using delayed_init = data_structures::delayed_initialization_wrapper<T>;

  delayed_init<R1> result_1_;
  delayed_init<R2> result_2_;
  delayed_init<F> function_;

  // Also uninitialized at first, only take the atomic write on the slow path.
  // The stealer will init it to 2 while stealing, the 'stolen' sync will then make sure
  // everyone sees the value in correct order.
  std::atomic<unsigned short> results_missing_{};
};

}
}
}

#endif //PLS_INTERNAL_SCHEDULING_CONTINUATION_H_
