
#ifndef PLS_TASK_MANAGER_H_
#define PLS_TASK_MANAGER_H_

#include <memory>

#include "pls/internal/data_structures/aligned_stack.h"

namespace pls {
namespace internal {
namespace scheduling {

// TODO: Remove forward references
class task;

/**
 * Handles management of tasks in the system. Each thread has a local task manager,
 * responsible for allocating, freeing and publishing tasks for stealing.
 *
 * The manager therefore acts as the deque found in work stealing, as well as the memory
 * management for the tasks (as both are deeply intertwined in our implementation to
 * integrate the memory management into the stealing procedure.
 */
class task_manager {
  using task_manager_offset = data_structures::aligned_stack::stack_offset;

 public:
  // Data each task needs to store to enable the 'return_task' functionality.
  using task_manager_state = task_manager_offset;

  // Construct a task onto the stack. Stores the previous offset in the newly constructed task.
  template<class T, typename ...ARGS>
  T *push_task(ARGS ...args);
  // Publishes a task on the stack, i.e. makes it visible for other threads to steal.
  void publish_task(task *task);
  // Return a no longer needed task to the stack. Must be the current most top task (will reset the stack pointer).
  void return_task(task *task);

  // Try to pop a local task from this task managers stack.
  task *pop_local_task();
  // Try to steal a task from a remote task_manager instance.
  // The returned task pointer is valid during the lifetyme of the task.
  // The returned task pointer must be returned to this task_manager instance.
  // (This is because we can either decide to just steal a remote task pointer or to copy the whole task)
  task *pop_remote_task(task_manager &other);

  explicit task_manager(data_structures::aligned_stack &task_stack) : task_stack_{task_stack} {}

 private:
  data_structures::aligned_stack &task_stack_;
};

template<size_t NUM_TASKS, size_t MAX_STACK_SIZE>
class static_task_manager {
 public:
  static_task_manager() : static_task_stack_{}, task_manager_{static_task_stack_} {};
  task_manager &get_task_manager() { return task_manager_; }

 private:
  data_structures::static_aligned_stack<MAX_STACK_SIZE> static_task_stack_;
  task_manager task_manager_;
};

}
}
}
#endif //PLS_TASK_MANAGER_H_
